#!/usr/bin/env python
# from pdb import set_trace
from sys import exit
import logging
from argparse import ArgumentParser, FileType
from bento_mdf.mdf import MDF
from bento_mdf.validator import MDFValidator

ap = ArgumentParser(description="Validate MDF against JSONSchema")
ap.add_argument('--schema',
                  help="MDF JSONschema file", type=FileType('r'),
                  dest="schema")
ap.add_argument('--quiet',
                  help="Suppress output; return only exit value",
                  action="store_true",
                  dest="quiet")
ap.add_argument('mdf_files',nargs='+',
                  metavar='mdf-file',
                  type=FileType('r'),
                  help="MDF yaml files for validation")
ap.add_argument("--log-file",
                help="Log file name")

def test(args, logger):
    retval = 0
    v = MDFValidator(args.schema, *args.mdf_files, logger=logger)
    if not v.load_and_validate_schema():
        retval += 1
    if not v.load_and_validate_yaml():
        retval += 1
    if not v.validate_instance_with_schema():
        retval += 1
    if not retval:
        for f in args.mdf_files:
            f.seek(0)
        if not MDF(*args.mdf_files, handle="test", logger=logger):
            retval += 1
    return retval


if __name__ == '__main__':
    args = ap.parse_args()
    logger = logging.getLogger("test-mdf")
    logger.setLevel(logging.DEBUG)
    fmt = logging.Formatter(fmt="%(asctime)s:%(name)s (%(levelname)s) - %(message)s")
    shdl = logging.StreamHandler()
    shdl.setLevel(logging.INFO)
    shdl.setFormatter(fmt)
    logger.addHandler(shdl)
    if args.quiet:
        shdl.setLevel(logging.CRITICAL)
    if args.log_file:
        fhdl = logging.FileHandler(args.log_file)
        fhdl.setLevel(logging.DEBUG)
        fhdl.setFormatter(fmt)
        logger.addHandler(fhdl)
    exit(test(args, logger))  # emit return val (0 = good) to os



import sys
import logging
import os.path
sys.path.append("..")
from bento_mdf.mdf import MDF
from bento_meta.objects import *
from warnings import warn
# NOTE: the diff class was changed from keeping the final data structure "result"
#       from being 'set' based to being 'list' so that it could be dumped into
#       json structure (which is incompatible with sets)


class Diff:
    """for manipulating the final result data structure when diff models"""

    def __init__(self):
        """sets holds tree of models, as it is parsed"""
        self.sets = {"nodes": {}, "edges": {}, "props": {}}
        # "terms": {} }
        self.clss = {"nodes": Node, "edges": Edge, "props": Property}
        """This will eventually hold the diff results"""
        self.result = {}

    def update_result(self, thing, entk, att, a_att, b_att):
        logging.info(
            "  entering update_result with thing {}, entk {}, att {}".format(
                thing, entk, att
            )
        )
        if thing not in self.result:
            self.result[thing] = {}
        if entk not in self.result[thing]:
            self.result[thing][entk] = {}
        if att not in self.result[thing][entk]:
            self.result[thing][entk][att] = {}
        cleaned_a_att = self.sanitize_empty(a_att)
        cleaned_b_att = self.sanitize_empty(b_att)
        self.result[thing][entk][att]["a"] = cleaned_a_att
        self.result[thing][entk][att]["b"] = cleaned_b_att

    def sanitize_empty(self, item):
        return self.sanitize_empty_list(item)

    def sanitize_empty_set(self, item):
        """an option to turn 'a':set() to 'a':None in final result"""
        if item != set():
            return item
        else:
            return None

    def sanitize_empty_list(self, item):
        """an option to turn 'a': [] to 'a':None in final result"""
        if item != list():
            return item
        else:
            return None

    def valuesets_are_different(self, vs_a, vs_b):
        """see if the group of terms in each value set is different"""

        # compare sets of terms
        # a_att.terms
        #   {'FFPE': <bento_meta.objects.Term object at 0x10..>,
        #    'Snap Frozen': <bento_meta.objects.Term object at 0x10..>}
        # set(a_att.terms)
        #   {'Snap Frozen', 'FFPE'}
        set_of_terms_in_a = set(vs_a.terms)
        set_of_terms_in_b = set(vs_b.terms)

        if set_of_terms_in_a == set_of_terms_in_b:
            return False
        else:
            return True

    def finalize_result(self):
        """adds info for uniq nodes, edges, props from self.sets back to self.result"""
        logging.info("finalizing result")
        for key, value in self.sets.items():
            logging.debug("key {} value {} ".format(key, value))
            logging.debug("sets is {}".format(self.sets))
            logging.debug("result is {}".format(self.result))

            if 0:
                if (value["a"] != set()) or (value["b"] != set()):
                    cleaned_a = self.sanitize_empty(value["a"])
                    cleaned_b = self.sanitize_empty(value["b"])

                    # the key (node/edges/prop) may not be in results (b/c no common diff yet found!)
                    if key not in self.result.keys():
                        self.result[key] = {}
                    self.result[key].update({"a": cleaned_a, "b": cleaned_b})

            if (value["a"] != list()) or (value["b"] != list()):
                cleaned_a = self.sanitize_empty(value["a"])
                cleaned_b = self.sanitize_empty(value["b"])

                # the key (node/edges/prop) may not be in results (b/c no common diff yet found!)
                if key not in self.result.keys():
                    self.result[key] = {}
                self.result[key].update({"a": cleaned_a, "b": cleaned_b})


def diff_models(mdl_a, mdl_b):
    """
    find the diff between two models
    populate the diff results into "sets" and keep some final stuff in result.result
    """
    diff_ = Diff()
    sets = diff_.sets
    clss = diff_.clss

    logging.info("point A")
    # set_trace()

    for thing in sets:
        aset = set(getattr(mdl_a, thing))
        bset = set(getattr(mdl_b, thing))
        sets[thing]["a"] = sorted(list(set(aset - bset)))
        sets[thing]["b"] = sorted(list(set(bset - aset)))
        sets[thing]["common"] = sorted(list(set(aset & bset)))

        logging.debug("ok, where is {} at?".format(thing))
        logging.debug("aset is {}".format(aset))
        logging.debug("bset is {}".format(bset))

        logging.debug(" you want a:{}".format(sets[thing]["a"]))
        logging.debug(" you want b:{}".format(sets[thing]["b"]))

    logging.info("point B")
    # set_trace()

    for thing in sets:
        logging.info("now doing ..{}".format(thing))
        # cls becomes a "Node" object, "Edge" object, etc
        cls = clss[thing]

        simple_atts = [x for x in cls.attspec_ if cls.attspec_[x] == "simple"]
        obj_atts = [x for x in cls.attspec_ if cls.attspec_[x] == "object"]
        coll_atts = [x for x in cls.attspec_ if cls.attspec_[x] == "collection"]

        for entk in sets[thing]["common"]:
            logging.info("...common entk is {}".format(entk))
            a_ent = getattr(mdl_a, thing)[entk]
            b_ent = getattr(mdl_b, thing)[entk]

            # try and see if the simple attributes are the same
            logging.info("...simple")
            for att in simple_atts:
                if getattr(a_ent, att) == getattr(b_ent, att):
                    logging.info("...comparing simple {}".format(getattr(a_ent, att)))
                    logging.info("...comparing simple {}".format(getattr(b_ent, att)))
                    continue
                else:
                    diff_.update_result(
                        thing,
                        entk,
                        att,
                        sorted(getattr(a_ent, att)),
                        sorted(getattr(b_ent, att)),
                    )

            # try and see if the "object" type is the same?
            #     a_att,b_att are things like "valuesets", "properties"
            logging.info("...object")
            for att in obj_atts:
                a_att = getattr(a_ent, att)
                b_att = getattr(b_ent, att)

                if a_att == b_att:  # only if both 'None' *or* is same object
                    continue
                if not a_att or not b_att:  # one is 'None'
                    diff_.update_result(thing, entk, att, sorted(a_att), sorted(b_att))
                    continue

                if type(a_att) == type(b_att):
                    if type(a_att) == ValueSet:  # kludge for ValueSet+Terms
                        if diff_.valuesets_are_different(a_att, b_att):
                            diff_.update_result(
                                thing,
                                entk,
                                att,
                                sorted(list(set(a_att.terms) - set(b_att.terms))),
                                sorted(list(set(b_att.terms) - set(a_att.terms))),
                            )
                    # items are something-other-than valuesets
                    elif getattr(a_att, "handle"):
                        if a_att.handle == b_att.handle:
                            continue
                        else:
                            diff_.update_result(
                                thing, entk, att, sorted(a_att), sorted(b_att)
                            )
                    else:
                        warn(
                            "can't handle attribute with type {}".format(
                                type(a_att).__name__
                            )
                        )
                        logging.warning(
                            "can't handle attribute with type {}".format(
                                type(a_att).__name__
                            )
                        )
                else:
                    diff_.update_result(thing, entk, att, sorted(a_att), sorted(b_att))

            # try and see if the "collection" set is the same?
            logging.info("...collection")
            for att in coll_atts:
                aset = set(getattr(a_ent, att))
                bset = set(getattr(b_ent, att))
                if aset != bset:
                    diff_.update_result(
                        thing,
                        entk,
                        att,
                        sorted(list(set(aset - bset))),
                        sorted(list(set(bset - aset))),
                    )

    logging.info("done")
    diff_.finalize_result()
    return diff_.result


if __name__ == "__main__":
    # m = hashlib.md5();
    # m.update(b"Hey dude");
    # print(m.hexdigest());

    smp_dir = "../tests/samples"
    mdl_a = MDF(os.path.join(smp_dir, "test-model-a.yml"), handle="test")
    mdl_b = MDF(os.path.join(smp_dir, "test-model-b.yml"), handle="test")
    result = diff_models(mdl_a.model, mdl_b.model)
    print(result)
    pass



#!/usr/bin/env python
import sys
sys.path.insert(0, '..')
import os
import getpass
import argparse
from bento_mdf.mdf import MDF
from bento_meta.mdb import WriteableMDB, make_nanoid, load_mdf

parser = argparse.ArgumentParser(description="Load model in MDF into an MDB")
parser.add_argument('files', nargs="*",
                    metavar="MDF-FILE", help="MDF file(s)/url(s)")
parser.add_argument('--commit', default='',
                    help="commit SHA1 for MDF instance (if any)",
                    required=True)
parser.add_argument('--handle', help="model handle")
parser.add_argument('--user', help="MDB username")
parser.add_argument('--passw', help="MDB password")
parser.add_argument('--bolt', metavar="BoltURL", help="MDB Bolt url endpoint (specify as 'bolt://...')")
parser.add_argument('--put', action='store_true', help="Load model to database")
parser.add_argument('--make-nanoids', action='store_true', help="Add new nanoids to graph nodes")
# example:
# args = parser.parse_args([
#     "https://raw.githubusercontent.com/CBIIT/icdc-model-tool/master/model-desc/icdc-model.yml",
#     "https://raw.githubusercontent.com/CBIIT/icdc-model-tool/master/model-desc/icdc-model-props.yml",
#     "--commit",
#     "a4aa9a43b9ad2087638ceaeef50d4a22a4b9b959",
#     "--handle",
#     "ICDC",
#     "--bolt",
#     "bolt://localhost:7687",
#     "--user",
#     "neo4j",
#     "--passw",
#     getpass.getpass()
#     ])

args = parser.parse_args()

if not args.files:
    parser.print_help()
    parser.exit(1)

if args.put and not args.passw:
    args.passw = getpass.getpass()

print("load model from MDFs")
mdf = MDF(*args.files, handle=args.handle, _commit=args.commit, raiseError=True)
model = mdf.model
if (args.bolt):
    mdb = WriteableMDB(uri=args.bolt, user=args.user, password=args.passw)
    model.mdb = mdb

if args.put:
    print("Put model to DB")
    load_mdf(mdf,model.mdb)
    if args.make_nanoids:
        print("Add nanoids to nodes")
        with mdb.driver.session() as s:
            result = s.run(
                "match (n {_commit:$commit}) where not exists(n.nanoid) "
                "with n limit 1 set n.nanoid=$nanoid return n",
                {"commit":args.commit, "nanoid":make_nanoid()}
            )
            while (result.peek()):
                result = s.run(
                    "match (n {_commit:$commit}) where not exists(n.nanoid) "
                    "with n limit 1 set n.nanoid=$nanoid return n",
                    {"commit":args.commit, "nanoid":make_nanoid()}
                    )



"""
bento_mdf.mdf
==============

This module contains :class:`MDF`, a class for reading a graph data model in
Model Description Format into a :class:`bento_meta.model.Model` object, and 
writing the opposite way.

"""
import sys
import yaml
import logging
from tqdm import tqdm
from urllib.parse import unquote
from tempfile import TemporaryFile
from bento_mdf.validator import MDFValidator
from bento_meta.model import Model
from bento_meta.entity import ArgError, CollValue
from bento_meta.objects import (
    Node,
    Edge,
    Property,
    Term,
    ValueSet,
    Tag,
    Origin,
)
import re
import requests
from collections import ChainMap
from nanoid import generate
import json

from pdb import set_trace

sys.path.extend([".", ".."])

def make_nano():
    return generate(
        size=6,
        alphabet="abcdefghijkmnopqrstuvwxyzABCDEFGHJKMNPQRSTUVWXYZ0123456789"
    )

class MDF(object):
    def __init__(self, *yaml_files, handle=None, model=None, _commit=None,
                 raiseError=False, logger=logging.getLogger(__name__)):
        """Create a :class:`Model` from MDF YAML files/Write a :class:`Model` to YAML
        :param str|file|url *yaml_files: MDF filenames or file objects, 
        in desired merge order
        :param str handle: Handle (name) for the resulting Model
        :param :class:`Model` model: Model to convert to MDF
        :param boolean raiseError: raise on error if True
        :param :class:`logging.Logger` logger: Python logger (suitable default)
        :attribute model: the :class:`bento_meta.model.Model` created"""
        if not model and (not handle or not isinstance(handle, str)):
            raise ArgError("arg handle= must be a str - name for model")
        if model and not isinstance(model,Model):
            raise ArgError("arg model= must be a Model instance")
            
        self.files = yaml_files
        self.schema = {}
        self._model = model
        self._commit = _commit
        self._terms = {}
        self._props = {}
        self.logger = logger
        if model:
            self.handle = model.handle
        else:
            self.handle = handle
        if self.files:
            self.load_yaml()
            self.create_model(raiseError=raiseError)
        else:
            if not model:
                self.logger.warning("No MDF files or model provided to constructor")

    @property
    def model(self):
        """The :class:`bento_meta.model.Model` object created from the 
           MDF input"""
        return self._model

    def load_yaml(self, verify=True):
        """Validate and load YAML files or open file handles specified in constructor"""
        vargs = []
        for f in self.files:
            if isinstance(f, str):
                if re.match("(?:file|https?)://", f):
                    response = requests.get(f, verify=verify)
                    if not response.ok:
                        self.logger.error(
                            "Fetching url {} returned code {}".format(
                                response.url, response.status_code
                            ))
                        raise ArgError(
                            "Fetching url {} returned code {}".format(
                                response.url, response.status_code
                            )
                        )
                    response.encoding = "utf8"
                    fh = TemporaryFile()
                    for chunk in response.iter_content(chunk_size=128):
                        fh.write(chunk)
                    fh.seek(0)
                    vargs.append(fh)
                else:
                    fh = open(f, "r")
                    vargs.append(fh)
            else:
                vargs.append(f)
                
        v = MDFValidator(None, *vargs, raiseError=True)
        self.schema = v.load_and_validate_yaml()

    def create_model(self, raiseError=False):
        """Create :class:`Model` instance from loaded YAML
        :param boolean raiseError: Raise if MDF errors found
        Note: This is brittle, since the syntax of MDF is hard-coded
        into this method."""
        success=True
        if not self.schema.keys():
            raise ValueError("attribute 'schema' not set - are yamls loaded?")
        if (self.handle):
            self._model = Model(handle=self.handle)
        elif (self.schema.get("Handle")):
            self.handle = self.schema["Handle"]
            self._model = Model(handle=self.schema["Handle"])
        else:
            self.logger.error("Model handle not present in MDF nor provided in args")
            success = False
            
        ynodes = self.schema["Nodes"]
        yedges = self.schema["Relationships"]
        ypropdefs = self.schema["PropDefinitions"]
        yunps = self.schema.get("UniversalNodeProperties")
        yurps = self.schema.get("UniversalRelationshipProperties")
        yterms = self.schema.get("Terms")
        if yterms:
            yterms = yterms.as_dict()

        # create terms first, if any -- properties depend on these
        if yterms:
            for t_hdl in tqdm(yterms):
                ytm = yterms[t_hdl]
                self.create_or_merge_term_from_mdf(ytm, t_hdl)
        
        # create nodes
        for n in ynodes:
            yn = ynodes[n]
            init = {"handle": n, "model": self.handle, "_commit": self._commit}
            if 'Desc' in yn and yn['Desc']:
                init['desc'] = yn['Desc']
            if 'NanoID' in yn and yn['NanoID']:
                init['nanoid'] = yn['NanoID']
            node = self._model.add_node(init)
            if "Tags" in yn:
                for t in yn["Tags"]:
                    node.tags[t] = Tag({"key": t,
                                        "value": yn["Tags"][t],
                                        "_commit": self._commit})
            if "Term" in yn:
                self.annotate_entity_from_mdf(node, yn["Term"])
                
        # create edges (relationships)
        for e in yedges:
            ye = yedges[e]
            for ends in ye["Ends"]:
                if ends["Src"] not in self._model.nodes:
                    self.logger.warning(
                        "No node '{src}' defined for edge "
                        "spec '{ename}' from '{src}' to '{dst}'"
                        .format(src=ends["Src"], dst=ends["Dst"],
                                ename=e)
                        )
                    continue
                if ends["Dst"] not in self._model.nodes:
                    self.logger.warning(
                        "No node '{dst}' defined for edge "
                        "spec '{ename}' from '{src}' to '{dst}'"
                        .format(src=ends["Src"], dst=ends["Dst"],
                                ename=e)
                        )
                    continue
                init = {
                    "handle": e,
                    "model": self.handle,
                    "src": self._model.nodes[ends["Src"]],
                    "dst": self._model.nodes[ends["Dst"]],
                    "multiplicity": ends.get("Mul")
                    or ye.get("Mul"),
                    "desc": ends.get("Desc") or ye.get("Desc"),
                    "_commit": self._commit
                }
                if not init["multiplicity"]:
                    self.logger.warning("edge '{ename}' from '{src}' to '{dst}' "
                                        "does not specify a multiplicity".
                                        format(ename=e, src=ends["Src"],
                                               dst=ends["Dst"]))
                    init["multiplicity"] = Edge.default("multiplicity")
                if init["multiplicity"] not in ('many_to_many', 'many_to_one',
                                                'one_to_many', 'one_to_one'):
                    self.logger.warning("edge '{ename}' from '{src}' to '{dst}'"
                                        " has non-standard multiplicity '{mult}'".
                                        format(ename=e, src=ends["Src"],
                                               dst=ends["Dst"], mult=init["multiplicity"]
                                   )
                    )
                edge = self._model.add_edge(init)
                Tags = ye.get("Tags") or ends.get("Tags")
                if Tags:
                    # tags = CollValue({}, owner=edge, owner_key="tags")
                    for t in Tags:
                        edge.tags[t] = Tag({"key": t,
                                            "value": Tags[t],
                                            "_commit": self._commit})
                yterm = ends.get("Term") or ye.get("Term")
                if yterm:
                    self.annotate_entity_from_mdf(edge, yterm)

        # create properties
        propnames = {}
        for ent in ChainMap(self._model.nodes, self._model.edges).values():
            if isinstance(ent, Node):
                pnames = ynodes[ent.handle]["Props"]
                if yunps:  # universal node props
                    pnames.extend(yunps["mayHave"] if yunps.get("mayHave") else [])
                    pnames.extend(yunps["mustHave"] if yunps.get("mustHave") else [])
                if pnames:
                    propnames[ent] = pnames
            elif isinstance(ent, Edge):
                # props elts appearing in Ends hash take precedence over 
                # Props elt in the handle's hash
                (hdl, src, dst) = ent.triplet
                ends = [
                    e
                    for e in yedges[hdl]["Ends"]
                    if e["Src"] == src and e["Dst"] == dst
                ]
                if len(ends) > 1:
                    self.logger.warning("edge '{ename}' has more than one Ends pair Src:'{src}',"
                                        "Dst:'{dst}'".format(ename=hdl, src=src, dst=dst))
                end = ends[0]
            
                # note the end-specified props _replace_ the edge-specified props,
                # they are not merged:
                pnames = end.get("Props") or yedges[hdl].get("Props")
                if yurps:  # universal relationship props
                    pnames.extend(yurps["mayHave"] if yurps.get("mayHave") else [])
                    pnames.extend(yurps["mustHave"] if yurps.get("mustHave") else [])
                if pnames:
                    propnames[ent] = pnames
            else:
                self.logger.error(
                    "Unhandled entity type {type} for properties".format(
                        type=type(ent).__name__
                    )
                )
                success = False
        prop_of = {}
        for ent in propnames:
            for p in propnames[ent]:
                if prop_of.get(p):
                    prop_of[p].append(ent)
                else:
                    prop_of[p] = [ent]
        defns_for = set(ypropdefs.keys())
        for pname in prop_of:
            for ent in prop_of[pname]:
                key = ent.handle+"."+pname
                ypdef = ypropdefs.get(key)
                if not ypdef:
                    key = pname
                    ypdef = ypropdefs.get(pname)
                if not ypdef:
                    self.logger.warning(
                        "property '{pname}' does not have a corresponding "
                        "propdef for entity '{handle}'".format(
                            pname=pname, handle=ent.handle
                        )
                    )
                    continue

                else:
                    if key in defns_for:
                        defns_for.remove(key)
                prop = self.create_or_merge_prop_from_mdf(ypdef, p_hdl=pname)
                self._model.add_prop(ent, prop)
                ent.props[prop.handle] = prop
        if defns_for:
            self.logger.warning(
                "No properties in model corresponding to the following "
                "PropDefintions: {}".format(defns_for))
        if raiseError and not success:
            raise RuntimeError("MDF errors found; see log output.")
        return self._model

    def create_or_merge_prop_from_mdf(self, ypdef, p_hdl):
        init = {"handle": p_hdl,
                "model": self.handle,
                "_commit": self._commit}
        if (init["model"], init["handle"]) in self._props:
            pass  # merge property
        else:
            if 'Desc' in ypdef and ypdef['Desc']:
                init['desc'] = ypdef['Desc']
            if 'NanoID' in ypdef and ypdef['NanoID']:
                init['nanoid'] = ypdef['NanoID']
            if 'Type' in ypdef:
                init.update(self.calc_value_domain(ypdef["Type"], p_hdl))
            elif 'Enum' in ypdef:
                init.update(self.calc_value_domain(ypdef["Enum"], p_hdl))
            else:
                self.logger.warning(
                    "property '{p_hdl}' does not "
                    "specify a data type".format(p_hdl=p_hdl)
                )
                init["value_domain"] = Property.default("value_domain")
            prop = Property(init)
            if "Tags" in ypdef:
                for t in ypdef["Tags"]:
                    prop.tags[t] = Tag({"key": t,
                                        "value": ypdef["Tags"][t],
                                        "_commit": self._commit})
            if "Term" in ypdef:
                self.annotate_entity_from_mdf(prop, ypdef["Term"])
            self._props[(init["model"], init["handle"])] = prop
        return self._props[(init["model"], init["handle"])]
        
    def create_or_merge_term_from_mdf(self, ytm, t_hdl=None):
        if not "Value" in ytm:
            self.logger.error(
                "Term specs must have a Value key and a non-null string value"
            )
            return False
        tm = {}
        tm["_commit"] = self._commit
        if 'Origin' in ytm:
            tm["origin_name"] = ytm["Origin"]
        else:
            self.logger.warning(
                "No Origin provided for term '{term}'".format(term=t_hdl)
            )
            tm["origin_name"] = self.handle
        tm["value"] = ytm["Value"]
        if (tm["value"],tm["origin_name"]) in self._terms:
            pass  # merge term
        else:
            if 'Definition' in ytm and ytm['Definition']:
                tm["origin_definition"] = unquote(ytm["Definition"])
            if 'Code' in ytm:
                tm["origin_id"] = ytm["Code"]
            if 'Version' in ytm:
                tm["origin_version"] = ytm["Version"]
            if 'Handle' in ytm:
                tm["handle"] = ytm["Handle"]
            if 'NanoID' in ytm:
                tm["nanoid"] = ytm["NanoID"]
            self._terms[(tm["value"],tm["origin_name"])] = Term(tm)
        return self._terms[(tm["value"],tm["origin_name"])]

    def annotate_entity_from_mdf(self, ent, yterm_list):
        for yterm in yterm_list:
            tm = self.create_or_merge_term_from_mdf(yterm)
            self._model.annotate(ent, tm)
            if self._commit:
                if not ent.concept._commit:
                    ent.concept._commit = self._commit

    def calc_value_domain(self, typedef, pname=None):
        if isinstance(typedef, dict):
            if typedef.get("pattern"):
                return {"value_domain": "regexp",
                        "pattern": typedef["pattern"]}
            elif typedef.get("units"):
                return {
                    "value_domain": typedef.get("value_type"),
                    "units": ";".join(typedef.get("units")),
                }
            elif typedef.get("item_type"):
                if (typedef["value_type"] == 'list'):
                    i_domain = self.calc_value_domain(typedef["item_type"])
                    ret = {"value_domain": "list",
                           "item_domain": i_domain["value_domain"]}
                    if i_domain.get("pattern"):
                        ret["pattern"] = i_domain["pattern"]
                    if i_domain.get("units"):
                        ret["units"] = i_domain["units"]
                    if i_domain.get("value_set"):
                        ret["value_set"] = i_domain["value_set"]
                    return ret
                else:
                    self.logger.warning(
                        "MDF type descriptor defines item_type, but value_type"
                        " is {}, not 'list' (property '{}')".format(typedef["value_type"],pname))
            elif not typedef:
                self.logger.warning("MDF type descriptor is null for property '{}'".format(pname))
            else:
                # punt
                self.logger.warning(
                    "MDF type descriptor unrecognized: json looks like {} (property '{}')".
                    format(json.dumps(typedef),pname)
                    )
                return {"value_domain": json.dumps(typedef)}
        elif isinstance(typedef, list):  # a valueset: create value set and term objs
            vs = ValueSet({"nanoid": make_nano(), "_commit": self._commit})
            vs.handle = self.handle + vs.nanoid
            # interpret boolean values as strings
            if (isinstance(typedef[0], str) and
                re.match("^(?:https?|bolt)://", typedef[0])):  # looks like url
                vs.url = typedef[0]
            else:  # an enum
                for t in typedef:
                    if isinstance(t, bool):  # stringify booleans in term context
                        t = "True" if t else "False"
                    if t not in self._terms:
                        self._terms[t] = Term({
                            "value": t,
                            "origin_name": self.handle,
                            "_commit": self._commit
                        })
                    vs.terms[self._terms[t].value] = self._terms[t]
            return {"value_domain": "value_set", "value_set": vs}
        elif isinstance(typedef, str):
            return {"value_domain": typedef}
        else:
            self.logger.warning("Applying default value domain to property '{}'".format(pname))
            return {"value_domain": Property.default("value_domain")}

    def write_mdf(self, model=None, file=None):
        """Write a :class:`Model` to a model description file (MDF)
        :param :class:`Model` model: Model to convert (if None, use the model attribute of the MDF object)
        :param str|file file: File name or object to write to (default is None; just return the MDF as dict)
        :returns: MDF as dict"""
        if not model:
            model = self.model
        mdf = {"Nodes":{},
               "Relationships":{},
               "PropDefinitions":{},
               "Terms":{},
               "Handle":model.handle}
        for nd in sorted(model.nodes):
            node = model.nodes[nd]
            mdf_node = {}
            mdf["Nodes"][nd] = mdf_node
            if node.tags:
                mdf_node["Tags"] = {}
                for t in node.tags:
                    mdf_node["Tags"][t] = node.tags[t].value
            if node.concept:
                if not node.concept.terms:
                    self.logger.warning("Node '{}' has associated concept but with no terms defined".format(node.handle))
                else:
                    mdf_node["Term"] = [{
                        "Value": tm.value,
                        "Definition": tm.origin_definition,
                        "Origin": tm.origin,
                        "Code": tm.origin_id,
                        } for tm in node.concept.terms.values()]

            mdf_node["Props"] = [prop for prop in sorted(node.props)]

            if node.nanoid:
                mdf_node["NanoID"] = node.nanoid
            if node.desc:
                mdf_node["Desc"] = node.desc
        # write props only in Ends object (no default set of properties)
        # MDF - if props are defined above the Ends object, this is the default set of properties
        # for any fully qualified edge without explicit properties. But if there are explict
        # properties applied to a fully qualified edge (i.e., in the Ends object), these are
        # the only properties for that edge.
        for rl in sorted(model.edges):
            edge = model.edges[rl]
            mdf_edge = {}
            ends = {}
            if edge.handle in mdf["Relationships"]:
                mdf_edge = mdf["Relationships"][edge.handle]
            else:
                mdf["Relationships"][edge.handle] = mdf_edge
            ends = {"Src": edge.src.handle,
                    "Dst": edge.dst.handle}
            if "Ends" in mdf_edge:
                mdf_edge["Ends"].append(ends)
            else:
                mdf_edge["Ends"] = [ends]
            if not "Mul" in mdf_edge:
                mdf_edge["Mul"] = edge.multiplicity or Edge.default("multiplicity")
            else:
                if mdf_edge["Mul"] != edge.multiplicity:
                    ends["Mul"] = edge.multiplicity
            if edge.tags:
                mdf_edge["Tags"] = {}
                for t in edge.tags:
                    mdf_edge["Tags"][t] = edge.tags[t].value
            if edge.is_required:
                ends["Req"] = True
            if edge.props:
                ends["Props"] = sorted(list(set(edge.props)))
            else:
                ends["Props"] = None
            if edge.nanoid:
                ends["NanoID"] = edge.nanoid
            if edge.desc:
                if not mdf_edge.get("Desc"):
                    mdf_edge["Desc"] = edge.desc
                else:
                    ends["Desc"] = edge.desc
            if edge.concept:
                if not edge.concept.terms:
                    self.logger.warning("Edge '{}' has associated concept but with no terms defined".format(node.handle))
                else:
                    mdf_edge["Term"] = [
                        {
                            "Value": tm.value,
                            "Definition": tm.origin_definition,
                            "Origin": tm.origin,
                            "Code": tm.origin_id,
                        } for tm in edge.concept.terms.values()]
        prnames = []
        props = {}
        for pr in model.props:
            prname = pr[len(pr)-1]
            prnames.append(prname)
            # if props.get(prname):
            #    self.logger.warning("Property name collision at {}".format(pr))
            if not prname in props:
                props[prname] = model.props[pr]
        for prname in sorted(prnames):
            prop = props[prname]
            mdf_prop = {}
            mdf["PropDefinitions"][prname] = mdf_prop
            if prop.tags:
                mdf_prop["Tags"] = {}
                for t in prop.tags:
                    mdf_prop["Tags"][t] = prop.tags[t].value
            if prop.value_domain == "value_set":
                mdf_prop["Enum"] = self.calc_prop_type(prop)
                for t in prop.terms:
                    # if t in mdf["Terms"]:
                    #    self.logger.warning("Term collision at {} (property {})".format(t, prop.handle))
                    if not t in mdf["Terms"]:
                        mdf["Terms"][t] = {
                            "Value": prop.terms[t].value,
                            "Definition": prop.terms[t].origin_definition,
                            "Origin": prop.terms[t].origin,
                            "Code": prop.terms[t].origin_id,
                        }
            else:
                mdf_prop["Type"] = self.calc_prop_type(prop)
            if prop.is_required:
                mdf_prop["Req"] = True
            if prop.nanoid:
                mdf_prop["NanoID"] = prop.nanoid
            if prop.desc:
                mdf_prop["Desc"] = prop.desc
            if prop.concept:
                if not prop.concept.terms:
                    self.logger.warning("Property '{}' has associated concept but with no terms defined".format(node.handle))
                else:
                    mdf_prop["Term"] = [{
                        "Value": tm.value,
                        "Definition": tm.origin_definition,
                        "Origin": tm.origin,
                        "Code": tm.origin_id,
                    } for tm in prop.concept.terms.values()]
        if file:
            fh = file
            if isinstance(file,str):
                fh = open(file, "w")
            yaml.dump(mdf, stream=fh, indent=4)
            
        return mdf

    def calc_prop_type(self,prop):
        if not prop.value_domain:
            return Property.default("value_domain")
        if prop.value_domain == "regexp":
            if not prop.pattern:
                self.logger.warning("Property {} has 'regexp' value domain, but no pattern specified".format(prop.handle))
                return {"pattern":"^.*$"}
            else:
                return {"pattern":prop.pattern}
        if prop.units:
            return {"value_type":prop.value_domain, "units":prop.units.split(';')}
        if prop.value_domain == "value_set":
            if not prop.value_set:
                self.logger.warning("Property {} has 'value_set' value domain, but value_set attribute is None".format(prop.handle))
                return "string"
            values = []
            for trm in sorted(prop.terms):
                values.append(trm)
            return values
        # otherwise 
        return prop.value_domain



import delfick_project.option_merge as om
import yaml
import logging
import requests
from jsonschema import (
  validate, ValidationError, SchemaError, RefResolutionError
  )
from jsonschema import Draft6Validator as d6
from yaml.parser import ParserError, ScannerError
from yaml.constructor import ConstructorError
from yaml.nodes import MappingNode, SequenceNode
from pdb import set_trace

MDFSCHEMA_URL = "https://github.com/CBIIT/bento-mdf/raw/master/schema/mdf-schema.yaml"

def construct_mapping(self, node, deep=False):
    if not isinstance(node, MappingNode):
        raise ConstructorError(
            None, None, "expected a mapping node, but found %s" % node.id,
            node.start_mark)
    mapping = {}

    for key_node, value_node in node.value:
        key = self.construct_object(key_node, deep=deep)
        try:
            hash(key)
        except TypeError as exc:
            raise ConstructorError("while constructing a mapping",
                                   node.start_mark,
                                   "found unacceptable key (%s)" %
                                   exc, key_node.start_mark)
        value = self.construct_object(value_node, deep=deep)
        if key in mapping:
            raise ConstructorError("while constructing a mapping",
                                   node.start_mark,
                                   "found duplicated key (%s)" %
                                   key, key_node.start_mark)
        mapping[key] = value
    return mapping


def construct_sequence(self, node, deep=False):
    if not isinstance(node, SequenceNode):
        raise ConstructorError(None, None,
                               "expected a sequence node, but found %s" %
                               node.id, node.start_mark)
    elts = set()
    for c in node.value:
        if isinstance(c.value, str):  # just check lists of strings
            if c.value in elts:
                raise ConstructorError("while constructing a sequence",
                                       node.start_mark,
                                       "found duplicated element (%s)" %
                                       c.value)
            else:
                elts.add(c.value)
    return [self.construct_object(child, deep=deep) for child in node.value]


class MDFValidator:
    """Class that encapsulates schema and YAML instance validation for the
    Bento Model Description Format. Use to check and load MDF YAML into 
    a python dict (see load_and_validate)."""
    def __init__(self, sch_file, *inst_files, raiseError=False,
                 logger=logging.getLogger(__name__)):
        self.schema = None
        self.instance = om.MergedOptions()
        self.sch_file = sch_file
        self.inst_files = inst_files
        self.yloader = yaml.loader.Loader
        self.yaml_valid = False
        self.logger = logger
        self.raiseError = raiseError

        # monkey patches to detect dup keys, elts
        self.yloader.construct_mapping = construct_mapping
        self.yloader.construct_sequence = construct_sequence
        
    def load_and_validate_schema(self):
        if self.schema:
            return self.schema
        if not self.sch_file:
            try:
                sch = requests.get(MDFSCHEMA_URL)
                sch.raise_for_status()
                self.sch_file = sch.text
            except Exception as e:
                self.logger.error("Error in fetching mdf-schema.yml: \n{e}".format(e=e))
                if self.raiseError:
                    raise e
                return
        elif isinstance(self.sch_file, str):
            try:
                self.sch_file = open(self. sch_file, "r")
            except IOError as e:
                self.logger.error(e)
                if self.raiseError:
                    raise e
                return
        else:
            pass
        try:
            self.logger.info("Checking schema YAML =====")
            self.schema = yaml.load(self.sch_file, Loader=self.yloader)
        except ConstructorError as ce:
            self.logger.error("YAML error in MDF Schema '{fn}':\n{e}".format(
                fn=self.sch_file.name, e=ce))
            if self.raiseError:
                raise
            return
                
        except ParserError as e:
            self.logger.error("YAML error in MDF Schema '{fn}':\n{e}".format(
                  fn=self.sch_file.name, e=e))
            if self.raiseError:
                raise e
            return
        except Exception as e:
            self.logger.error("Exception in loading MDF Schema yaml: {}".format(e))
            if self.raiseError:
                raise e
            return
        self.logger.info("Checking as a JSON schema =====")
        try:
            d6.check_schema(self.schema)
        except SchemaError as se:
            self.logger.error("MDF Schema error: {}".format(se))
            if self.raiseError:
                raise se
            return
        except Exception as e:
            self.logger.error("Exception in checking MDF Schema: {}".format(e))
            if self.raiseError:
                raise e
            return
        return self.schema

    def load_and_validate_yaml(self):
        if self.instance:
            return self.instance
        if (self.inst_files):
            self.logger.info("Checking instance YAML =====")
            for inst_file in self.inst_files:
                if isinstance(inst_file, str):
                    inst_file = open(inst_file, "r")
                try:
                    inst_yaml = yaml.load(inst_file, Loader=self.yloader)
                    self.instance.update(inst_yaml)
                except ConstructorError as ce:
                    self.logger.error("YAML error in '{fn}':\n{e}".format(
                        fn=inst_file.name, e=str(ce)))
                    if self.raiseError:
                        raise ce
                    return
                except ParserError as e:
                    self.logger.error("YAML error in '{fn}':\n{e}".format(fn=inst_file.name,e=e))
                    if self.raiseError:
                        raise e
                    return
                except ScannerError as e:
                    self.logger.error("YAML error in '{fn}':\n{e}".format(fn=inst_file.name,e=e))
                    if self.raiseError:
                        raise e
                    return
                except Exception as e:
                    self.logger.error("Exception in loading yaml (instance): {}".format(e))
                    if self.raiseError:
                        raise e
                    return
        return self.instance
        self.logger.error("No instance yaml(s) specified")
        return None
    
    def validate_instance_with_schema(self):
        if not self.schema:
            self.logger.warning("No valid schema; skipping this validation")
            return
        if not self.instance:
            self.logger.warning("No valid yaml instance; skipping this validation")
            return
        if (self.instance):
            self.logger.info("Checking instance against schema =====")
            try:
                validate(instance=self.instance.as_dict(), schema=self.schema)
            except ConstructorError as ce:
                self.logger.error(ce)
                if self.raiseError:
                    raise ce
                return
            except RefResolutionError as re:
                self.logger.error(re)
                if self.raiseError:
                    raise re
                return
            except ValidationError as ve:
                for e in d6(self.schema).iter_errors(self.instance.as_dict()):
                    self.logger.error(e)
                if self.raiseError:
                    raise ve
                return
            except Exception as e:
                self.logger.error("Exception during validation: {}".format(e))
                if self.raiseError:
                    raise e
                return
        return self.instance



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['bento_mdf']

package_data = \
{'': ['*']}

install_requires = \
['bento-meta>=0.0.50,<0.0.51',
 'delfick-project>=0.7.9,<0.8.0',
 'jsonschema>=4.17.3,<5.0.0',
 'pyyaml>=6.0,<7.0',
 'requests>=2.28.2,<3.0.0',
 'tqdm>=4.64.1,<5.0.0']

scripts = \
['bin/test-mdf.py', 'bin/load-mdf.py']

setup_kwargs = {
    'name': 'bento-mdf',
    'version': '0.4.3',
    'description': 'Python driver/validator for Bento Model Description Format',
    'long_description': 'bento_mdf\n=======\n\nPython 3 drivers for the graph [Model Description Format](https://github.com/CBIIT/bento-mdf)\n\nThis directory provides ``test-mdf.py``, a standalone command line MDF validator.\n\n## Installation\n\nInstall the latest version (including scripts below) from GitHub using\nan up-to-date pip:\n\n\tpip install bento_mdf@git+https://github.com/CBIIT/bento-mdf.git#egg=subdir\\&subdirectory=drivers/python\n\n## Scripts\n\nScripts [`test-mdf.py`](./test-mdf.py) and\n[`load-mdf.py`](./load-mdf.py) are included in the\ndistribution. `test-mdf` is a verbose validator that can be used to\nfind issues in a set of local MDFs using the [MDF\nJSONSchema](../../schema/mdf-schema.yaml). `load-mdf` will load a\nvalid set of MDFs into an existing [Neo4j](https://neo4j.com) [Metamodel Database](https://github.com/CBIIT/bento-meta).\n\n\n## `test-mdf` Usage\n\n    $ test-mdf.py -h\n    usage: test-mdf.py [-h] [--schema SCHEMA] [--quiet] [--log-file LOG_FILE]\n                       mdf-file [mdf-file ...]\n\n    Validate MDF against JSONSchema\n\n    positional arguments:\n      mdf-file             MDF yaml files for validation\n\n    optional arguments:\n      -h, --help           show this help message and exit\n      --schema SCHEMA      MDF JSONschema file\n      --quiet              Suppress output; return only exit value\n      --log-file LOG_FILE  Log file name\n\nSee "Validator Notes" below.\n\n## `load-mdf` Usage\n\n    $ ./load-mdf.py -h\n    usage: load-mdf.py [-h] --commit COMMIT [--handle HANDLE] [--user USER] [--passw PASSW]\n                       [--bolt BoltURL] [--put]\n                       [MDF-FILE ...]\n\n    Load model in MDF into an MDB\n\n    positional arguments:\n      MDF-FILE         MDF file(s)/url(s)\n\n    optional arguments:\n      -h, --help       show this help message and exit\n      --commit COMMIT  commit SHA1 for MDF instance (if any)\n      --handle HANDLE  model handle\n      --user USER      MDB username\n      --passw PASSW    MDB password\n      --bolt BoltURL   MDB Bolt url endpoint (specify as \'bolt://...\')\n      --put            Load model to database\n\n## Validator `test-mdf.py`Notes\n\nThe ``--schema`` argument is optional. ``test-mdf.py`` will automatically retrieve the latest [mdf-schema.yaml](../../schema/mdf-schema.yaml) in the master branch of [this repo](https://github.com/CBIIT/bento-mdf).\n\nThe script tests both the syntax of the YAML (for both schema and MDF files), and the validity of the files with respect to the JSONSchema (for both schema and MDF files).\n\nThe errors are as emitted from the [PyYaml](https://pyyaml.org/wiki/PyYAMLDocumentation) and [jsonschema](https://python-jsonschema.readthedocs.io/en/stable/) packages, and can be rather obscure.\n\n* Successful test\n\n        $ test-mdf.py samples/ctdc_model_file.yaml samples/ctdc_model_properties_file.yaml \n        Checking schema YAML =====\n        Checking as a JSON schema =====\n        Checking instance YAML =====\n        Checking instance against schema =====\n\n* Bad YAML syntax\n\n        $ test-mdf.py samples/ctdc_model_bad.yaml samples/ctdc_model_properties_file.yaml \n        Checking schema YAML =====\n        Checking as a JSON schema =====\n        Checking instance YAML =====\n        YAML error in \'samples/ctdc_model_bad.yaml\':\n        while parsing a block mapping\n          in "samples/ctdc_model_bad.yaml", line 1, column 1\n        expected <block end>, but found \'<block mapping start>\'\n          in "samples/ctdc_model_bad.yaml", line 3, column 3\n\n* Schema-invalid YAML\n\n        $ test-mdf.py samples/ctdc_model_file_invalid.yaml samples/ctdc_model_properties_file.yaml \n        Checking schema YAML =====\n        Checking as a JSON schema =====\n        Checking instance YAML =====\n        Checking instance against schema =====\n        [\'show_node\', \'specimen_id\', \'biopsy_sequence_number\', \'specimen_type\'] is not of type \'object\'\n        \n        Failed validating \'type\' in schema[\'properties\'][\'Nodes\'][\'additionalProperties\']:\n            {\'$id\': \'#nodeSpec\',\n             \'properties\': {\'Category\': {\'$ref\': \'#/defs/snake_case_id\'},\n                            \'Props\': {\'oneOf\': [{\'items\': {\'$ref\': \'#/defs/snake_case_id\'},\n                                                 \'type\': \'array\',\n                                                 \'uniqueItems\': True},\n                                                {\'type\': \'null\'}]},\n                            \'Tags\': {\'$ref\': \'#/defs/tagsSpec\'}},\n             \'required\': [\'Props\'],\n             \'type\': \'object\'}\n        \n        On instance[\'Nodes\'][\'specimen\']:\n            [\'show_node\', \'specimen_id\', \'biopsy_sequence_number\', \'specimen_type\']\n\n## Testing the tester\n\nThe validator code itself can be tested as follows:\n\n    pip install tox\n    cd bento-mdf/validators/mdf-validate\n    tox\n\n\n\n\n',
    'author': 'Mark A. Jensen',
    'author_email': 'mark.jensen@nih.gov',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'scripts': scripts,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
